local FILENAME = "custom_color.txt"

local MYCHAR_X = 130
local MYCHAR_Y = 70
local MYCHAR_SCALE = FRACUNIT

local PLAYERS_X = 215
local PLAYERS_Y = 55
local PLAYERS_SPACING_X = 60
local PLAYERS_SPACING_Y = 60
local PLAYERS_CURRENT_X
local PLAYERS_CURRENT_Y
local PLAYERS_SCALE = FRACUNIT*3/4
local PLAYERS_ROTSPEED = 2

local RAMP_X = 20
local RAMP_Y = 5
local RAMP_SLOT_WIDTH = 24
local RAMP_SLOT_HEIGHT = 12

local NAME_X = RAMP_X + RAMP_SLOT_WIDTH + 5
local NAME_Y = 5
local NAME_WIDTH = 130
local NAME_HEIGHT = 12
local NAME_LEN_MAX = 20

local MENUITEMS_X = RAMP_X + RAMP_SLOT_WIDTH + 5
local MENUITEMS_Y = 30
local MENUITEMS_WIDTH = 40
local MENUITEMS_HEIGHT = 12
local MENUITEMS = {"Save", "Exit"}

local PALETTE_X = RAMP_X + RAMP_SLOT_WIDTH + 5
local PALETTE_Y = 85
local PALETTE_SLOT_SIZE = 7

local mouse_x = 160 * FRACUNIT
local mouse_y = 120 * FRACUNIT
local ramp_selected = 0
local ramp_hover = -1
local palette_hover = -1
local button_hover = -1
local name_hover = false

local name_editing = false
local color_name = ""

local function clearColor(color, accessible)
	skincolors[color] = {
		name = "",
		ramp = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15},
		invcolor = SKINCOLOR_NONE,
		invshade = 0,
		chatcolor = 0,
		accessible = accessible
	}
end

for i = 1, 32
	local color_name = "SKINCOLOR_CUSTOM_" + i
	freeslot(color_name)
	local col = _G[color_name]
	clearColor(col, false)
	skincolors[col].name = "MyCustomColor" + i
end

local function getColor(player)
	return SKINCOLOR_CUSTOM_1 + #player
end

local function getColorData(player)
	return skincolors[getColor(player)]
end

local function isAlphanumeric(str)
	if not str:find("%W") then  -- Check if the string only contains numbers and letters
		if not str:find("^%d+$") then  -- Check if the string does not contain only numbers
			return true
		end
	end
	return false
end


local function drawPlayer(v, c, scale, skin, sprite2, frame_duration, num_frames, flip)
	local angle = (leveltime * PLAYERS_ROTSPEED) % 360
	local rotation = 1 + (((angle+22) / 45) % 8)
	local flags = 0
	if flip and rotation > 4
		flags = V_FLIP
	end
	local colormap = v.getColormap(skin, c)
	
	if skin == "tails" and not (rotation >= 4 and rotation <= 6)
		local frame = ((leveltime/2) % 4)
		local patch = v.getSprite2Patch(skin, SPR2_TAL7, false, frame, rotation)
		local xx = sin(angle * ANG1) * 3
		v.drawScaled(PLAYERS_CURRENT_X * FRACUNIT + xx, PLAYERS_CURRENT_Y * FRACUNIT, scale, patch, flags, colormap)
	end
	
	local frame = ((leveltime/frame_duration) % num_frames)
	local patch = v.getSprite2Patch(skin, sprite2, false, frame, rotation)
	v.drawScaled(PLAYERS_CURRENT_X * FRACUNIT, PLAYERS_CURRENT_Y * FRACUNIT, scale, patch, flags, colormap)
	
	if skin == "tails" and rotation >= 4 and rotation <= 6
		local frame = ((leveltime/2) % 4)
		local patch = v.getSprite2Patch(skin, SPR2_TAL7, false, frame, rotation)
		local xx = sin(angle * ANG1) * 3
		v.drawScaled(PLAYERS_CURRENT_X * FRACUNIT + xx, PLAYERS_CURRENT_Y * FRACUNIT, scale, patch, flags, colormap)
	end
	
	PLAYERS_CURRENT_X = $ + PLAYERS_SPACING_X
end

local function drawPlayers(v, x, y)
	local x = PLAYERS_X
	local y = PLAYERS_Y
	local s = PLAYERS_SCALE
	local c = getColor(consoleplayer)
	
	PLAYERS_CURRENT_X = x
	PLAYERS_CURRENT_Y = y
	drawPlayer(v, c, s, "sonic",		SPR2_STND,	1, 1, true)
	drawPlayer(v, c, s, "tails",		SPR2_FLY,	2, 2, true)
	PLAYERS_CURRENT_X = x
	PLAYERS_CURRENT_Y = $ + PLAYERS_SPACING_Y
	drawPlayer(v, c, s, "knuckles",		SPR2_GLID,	3, 2, true)
	drawPlayer(v, c, s, "amy",			SPR2_MLEE,	4, 4, true)
	PLAYERS_CURRENT_X = x
	PLAYERS_CURRENT_Y = $ + PLAYERS_SPACING_Y
	drawPlayer(v, c, s, "fang",			SPR2_FIRE,	3, 4, false)
	drawPlayer(v, c, s, "metalsonic",	SPR2_WAIT,	16, 2, false)
end

local function drawMyChar(v, x, y)
	local c = getColor(consoleplayer)
	local skin = consoleplayer.skin
	local patch = v.getSprite2Patch(skin, SPR2_SIGN, false, 0, 0)
	v.drawScaled(MYCHAR_X * FRACUNIT, MYCHAR_Y * FRACUNIT, MYCHAR_SCALE, patch, flags, v.getColormap(skin, c))
end

local function drawPalette(v)
	local x = PALETTE_X
	local y = PALETTE_Y
	local color_data = getColorData(consoleplayer)
	local mx = mouse_x / FRACUNIT
	local my = mouse_y / FRACUNIT
	palette_hover = -1
	for i = 0, 254
		local slot_x = (i % 16)
		local slot_y = i / 16
		local xx = slot_x * PALETTE_SLOT_SIZE
		local yy = slot_y * PALETTE_SLOT_SIZE
		local x1 = x + xx
		local y1 = y + yy
		local x2 = x1 + PALETTE_SLOT_SIZE
		local y2 = y1 + PALETTE_SLOT_SIZE
		
		v.drawFill(x1, y1, PALETTE_SLOT_SIZE, PALETTE_SLOT_SIZE, i)
		
		if mx >= x1 and mx < x2 and my >= y1 and my < y2
			palette_hover = i
		end
	end
	
	local palette_selected = color_data.ramp[ramp_selected]
	local palette_selected_x = (palette_selected % 16)
	local palette_selected_y = palette_selected / 16
	local patch = v.cachePatch("PCURSOR")
	v.draw(x + palette_selected_x * PALETTE_SLOT_SIZE, y + palette_selected_y * PALETTE_SLOT_SIZE, patch)
	
	if palette_hover != -1
		local palette_hover_x = (palette_hover % 16)
		local palette_hover_y = palette_hover / 16
		local patch = v.cachePatch("PCURSOR2")
		v.draw(x + palette_hover_x * PALETTE_SLOT_SIZE, y + palette_hover_y * PALETTE_SLOT_SIZE, patch)
	end
end

local function drawRamp(v)
	local x = RAMP_X
	local y = RAMP_Y
	local color_data = getColorData(consoleplayer)
	local mx = mouse_x / FRACUNIT
	local my = mouse_y / FRACUNIT
	ramp_hover = -1
	for i = 0, 15
		local yy = i * RAMP_SLOT_HEIGHT
		local x1 = x
		local y1 = y + yy
		local x2 = x1 + RAMP_SLOT_WIDTH
		local y2 = y1 + RAMP_SLOT_HEIGHT
		local padding = 0
		
		if mx >= x1 and mx < x2 and my >= y1 and my < y2
			padding = 2
			ramp_hover = i
		end
		
		v.drawFill(x1 - padding, y1, RAMP_SLOT_WIDTH + padding * 2, RAMP_SLOT_HEIGHT, color_data.ramp[i])
		v.drawString(x1 + 2, y1 + 2, color_data.ramp[i], 0, "thin")
	end
	
	local patch = v.cachePatch("RCURSOR")
	v.drawScaled(x * FRACUNIT, (y + ramp_selected * RAMP_SLOT_HEIGHT) * FRACUNIT, FRACUNIT, patch)
end

local function drawButtons(v)
	local x = MENUITEMS_X
	local y = MENUITEMS_Y
	button_hover = -1
	for i = 0, #MENUITEMS - 1
		local mx = mouse_x / FRACUNIT
		local my = mouse_y / FRACUNIT
		local yy = i * MENUITEMS_HEIGHT
		local x1 = x
		local y1 = y + yy
		local x2 = x1 + MENUITEMS_WIDTH
		local y2 = y1 + MENUITEMS_HEIGHT
		local padding = 0
		
		if mx >= x1 and mx < x2 and my >= y1 and my < y2
			padding = 2
			button_hover = i
		end
		
		v.drawFill(x1 - padding, y1, MENUITEMS_WIDTH + padding * 2, MENUITEMS_HEIGHT, i * 16)
		v.drawString(x1 + 2, y1 + 2, MENUITEMS[i+1], 0, "thin")
	end
end

local function drawName(v)
	local mx = mouse_x / FRACUNIT
	local my = mouse_y / FRACUNIT
	local x1 = NAME_X
	local y1 = NAME_Y
	local x2 = x1 + NAME_WIDTH
	local y2 = y1 + NAME_HEIGHT
	local padding = 0
	
	name_hover = false
	if mx >= x1 and mx < x2 and my >= y1 and my < y2
		padding = 2
		name_hover = true
	end
	
	v.drawFill(x1 - padding, y1, NAME_WIDTH + padding * 2, NAME_HEIGHT, 170)
	
	local name = color_name
	if name_editing and (leveltime % 10 < 5)
		name = $ + "_"
	end
	v.drawString(x1 + 2, y1 + 2, name, V_ALLOWLOWERCASE, "thin")
end

local function drawCursor(v)
	if not consoleplayer or isdedicatedserver return end
	if not consoleplayer.colmenu_open return end
	local patch = v.cachePatch("MCURSOR")
	v.drawScaled(mouse_x, mouse_y, FRACUNIT/2, patch)
end

local function openMenu(player)
	player.colmenu_open = true
	player.colmenu_lock_angle = player.mo.angle
	player.colmenu_lock_cmd_aiming = player.cmd.aiming
	player.colmenu_lock_aiming = player.aiming
	if player == consoleplayer
		ramp_selected = 0
	end
	if not getColorData(player).accessible
		clearColor(getColor(player), true)
		getColorData(player).name = player.name
	end
	
	if not (gametyperules & GTR_TEAMS)
		COM_BufInsertText(player, "color "+getColor(player))
		player.skincolor = getColor(player)
		if player.mo and player.mo.valid
			player.mo.color = getColor(player)
		end
	end
	S_StartSound(nil, sfx_kc5e, player)
end

local function closeMenu(player)
	player.colmenu_open = nil
	player.colmenu_lock_angle = nil
	player.colmenu_lock_cmd_aiming = nil
	player.colmenu_lock_aiming = nil
	player.colmenu_cooldown = 2
	S_StartSound(nil, sfx_kc65, consoleplayer)
end

COM_AddCommand("customcolor", function(player)
	if (player.colmenu_open)
		closeMenu(player)
	elseif not player.colmenu_cooldown
		openMenu(player)
	end
end)

COM_AddCommand("customcolor_change_value", function(player, ramp_slot, color)
	getColorData(player).ramp[ramp_slot] = color
end)

COM_AddCommand("customcolor_change_all", function(player, ...)
	local ramp = {...}
	if ramp == nil or #ramp != 16
		if player == consoleplayer
			print("ERROR: please supply 16 colors for your ramp")
		end
		return
	end
	local data = getColorData(player)
	for i = 1, #ramp
		local col = tonumber(ramp[i])
		if col != nil
			data.ramp[i - 1] = col
		end
	end
	data.accessible = true
end)

COM_AddCommand("customcolor_change_name", function(player, name)
	if name and string.len(name) < NAME_LEN_MAX
		local data = getColorData(player)
		if not isAlphanumeric(name)
			print("Invalid skin name")
			color_name = data.name
			return
		end
		for c = 1, #skincolors - 1
			local color = skincolors[c]
			if string.lower(color.name) == string.lower(name)
				print("Skin name " + color.name + " already exists.")
				color_name = data.name
				return
			end
		end
		data.name = name
		data.accessible = true
		color_name = name
	end
end)

COM_AddCommand("customcolor_save", function(player)
	local color = getColor(player)
	local color_data = getColorData(player)
	local file = io.openlocal("client/" + FILENAME, "w")
	if not file
		print("ERROR: Could not create luafiles/client/" + FILENAME)
		return
	end
	
	file:write(color_data.name + "\n")
	
	local text = ""
	for i = 0, 15
		file:write(color_data.ramp[i])
		if i < 15
			file:write(" ")
		end
	end
	file:close()
	
	print("Saved your custom color to luafiles/client/" + FILENAME)
end, COM_LOCAL)

COM_AddCommand("customcolor_load", function(player, silent)
	local file = io.openlocal("client/" + FILENAME, "r")
	if not file
		if not silent
			print("You don't have a saved color yet!")
		end
		return
	end
	
	local name = file:read()
	local ramp = file:read()
	if name and ramp
		COM_BufInsertText(consoleplayer, "customcolor_change_name " + name)
		COM_BufInsertText(consoleplayer, "customcolor_change_all " + ramp)
		print("Loaded a custom color!")
	else
		print("Error: " + FILENAME + " data is invalid")
	end
end, COM_LOCAL)

addHook("PlayerQuit", function(player)
	getColorData(player).accessible = false
end)

addHook("PreThinkFrame", do
	for player in players.iterate
		if player.colmenu_open
			player.mo.angle = player.colmenu_lock_angle
			player.cmd.buttons = 0
			player.cmd.forwardmove = 0
			player.cmd.sidemove = 0
			player.cmd.aiming = player.colmenu_lock_cmd_aiming
			player.aiming = player.colmenu_lock_aiming
		end
		if player.colmenu_cooldown
			player.colmenu_cooldown = $ - 1
		end
	end
	if not consoleplayer or isdedicatedserver return end
	if not consoleplayer.colmenu_open return end
	mouse_x = $ + mouse.dx * 3000
	mouse_y = $ + mouse.dy * 3000
	mouse_x = max($, 0)
	mouse_x = min($, 320 * FRACUNIT)
	mouse_y = max($, 0)
	mouse_y = min($, 200 * FRACUNIT)
end)

addHook("KeyDown", function(keyevent)
	if not consoleplayer or isdedicatedserver return end
	if consoleplayer.colmenu_open
		if name_editing
			if (keyevent.name == "enter")
				name_editing = false
				COM_BufInsertText(consoleplayer, "customcolor_change_name " + color_name)
				return true
			end
			if (keyevent.name == "backspace")
				if string.len(color_name) > 0
					color_name = string.sub($, 1, string.len($) - 1)
					S_StartSound(nil, sfx_kc43, consoleplayer)
				end
				return true
			end
			if string.len(color_name) < NAME_LEN_MAX
				if (keyevent.num >= 199 and keyevent.num <= 201)
				or (keyevent.num >= 203 and keyevent.num <= 205)
				or (keyevent.num >= 207 and keyevent.num <= 210) //numpad fuck you
					local c
					if keyevent.num >= 199 and keyevent.num <= 201
						c = keyevent.num - 192
					elseif keyevent.num >= 203 and keyevent.num <= 205
						c = keyevent.num - 199
					elseif keyevent.num >= 207 and keyevent.num <= 209
						c = keyevent.num - 206
					else
						c = 0
					end
					color_name = $ .. tostring(c)
					S_StartSound(nil, sfx_kc50, consoleplayer)
					return true
				end
				if keyevent.num >= 48 and keyevent.num <= 57 //0 to 9
					local c = tostring(keyevent.num - 48)
					color_name = $ .. c
					S_StartSound(nil, sfx_kc50, consoleplayer)
					return true
				end
				if keyevent.num >= 97 and keyevent.num <= 122 //a to z
					local c = string.char(keyevent.num)
					if gamekeydown[182] or gamekeydown[183]
						c = string.upper($)
					end
					color_name = $ .. c
					S_StartSound(nil, sfx_kc50, consoleplayer)
					return true
				end
			end
			if keyevent.name == "mouse1" and not name_hover
				name_editing = false
				COM_BufInsertText(consoleplayer, "customcolor_change_name " + color_name)
			end
		end
		if keyevent.name == "mouse1"
			if name_hover
				name_editing = true
			elseif ramp_hover != -1
				ramp_selected = ramp_hover
				S_StartSound(nil, sfx_kc50, consoleplayer)
			elseif palette_hover != -1
				COM_BufInsertText(consoleplayer, "customcolor_change_value " + ramp_selected + " " + palette_hover)
				S_StartSound(nil, sfx_kc56, consoleplayer)
			elseif button_hover != -1
				if button_hover == 0
					COM_BufInsertText(consoleplayer, "customcolor_save")
					S_StartSound(nil, sfx_kc48, consoleplayer)
				elseif button_hover == 1
					COM_BufInsertText(consoleplayer, "customcolor")
				end
			end
			return true
		elseif (keyevent.name == "up arrow" or keyevent.name == "wheel 1 up") and not name_editing
			ramp_selected = $ - 1
			if ramp_selected < 0
				ramp_selected = 15
			end
			S_StartSound(nil, sfx_kc50, consoleplayer)
			return true
		elseif (keyevent.name == "down arrow" or keyevent.name == "wheel 1 down" or keyevent.name == "tab") and not name_editing
			ramp_selected = ($ + 1) % 16
			S_StartSound(nil, sfx_kc50, consoleplayer)
			return true
		elseif not name_editing
			COM_BufInsertText(consoleplayer, "customcolor")
			if keyevent.name == "escape"
				return true
			end
		end
	end
end)

addHook("PlayerSpawn", function(player)
	if getColorData(player).accessible
		if not (gametyperules & GTR_TEAMS)
			COM_BufInsertText(player, "color "+getColor(player))
			player.skincolor = getColor(player)
			if player.mo and player.mo.valid
				player.mo.color = getColor(player)
			end
		end
	else
		COM_BufInsertText(player, "customcolor_load 1")
		if not (gametyperules & GTR_TEAMS)
			COM_BufInsertText(player, "color "+getColor(player))
			player.skincolor = getColor(player)
			if player.mo and player.mo.valid
				player.mo.color = getColor(player)
			end
		end
	end
end)

hud.add(function(v)
	if not (consoleplayer and consoleplayer.colmenu_open) return end
	v.drawStretched(-32 * FRACUNIT, -8 * FRACUNIT, 600 * FRACUNIT, 400 * FRACUNIT, v.cachePatch("BPIXEL"), V_50TRANS)
	drawPlayers(v)
	drawPalette(v)
	drawMyChar(v)
	drawRamp(v)
	drawName(v)
	drawButtons(v)
	drawCursor(v)
end, "game")